from typing import Literal

from . import dartrs


LengthTag = (
    Literal["<|length:very_short|>"]
    | Literal["<|length:short|>"]
    | Literal["<|length:medium|>"]
    | Literal["<|length:long|>"]
    | Literal["<|length:very_long|>"]
)

AspectRatioTag = (
    Literal["<|aspect_ratio:ultra_wide|>"]
    | Literal["<|aspect_ratio:wide|>"]
    | Literal["<|aspect_ratio:square|>"]
    | Literal["<|aspect_ratio:tall|>"]
    | Literal["<|aspect_ratio:ultra_tall|>"]
)

RatingTag = (
    Literal["<|rating:sfw|>"]
    | Literal["<|rating:general|>"]
    | Literal["<|rating:sensitive|>"]
    | Literal["<|rating:nsfw|>"]
    | Literal["<|rating:questionable|>"]
    | Literal["<|rating:explicit|>"]
)

IdentityTag = (
    Literal["<|identity:none|>"]
    | Literal["<|identity:lax|>"]
    | Literal["<|identity:strict|>"]
)


def compose_prompt(
    prompt: str = "",
    copyright: str = "",
    character: str = "",
    rating: RatingTag = "<|rating:general|>",
    aspect_ratio: AspectRatioTag = "<|aspect_ratio:tall|>",
    length: LengthTag = "<|length:medium|>",
    identity: IdentityTag = "<|identity:none|>",
):
    return dartrs.compose_prompt_v2(
        copyright=copyright,
        character=character,
        rating=dartrs.RatingTag(rating),
        aspect_ratio=dartrs.AspectRatioTag(aspect_ratio),
        length=dartrs.LengthTag(length),
        identity_level=dartrs.IdentityTag(identity),
        prompt=prompt,
    )


class V2Model:
    model: dartrs.DartV2Mistral | dartrs.DartV2Mixtral

    def __init__(
        self,
        model: dartrs.DartV2Mistral | dartrs.DartV2Mixtral,
    ) -> None:
        self.model = model

    def generate(self, config: dartrs.GenerationConfig) -> str:
        return self.model.generate(config)


class MixtralModel(V2Model):
    @classmethod
    def from_pretrained(
        cls,
        hub_name: str,
        revision: str | None = None,
        dtype: dartrs.DartDType = dartrs.DartDType.FP32,
        device: dartrs.DartDevice = dartrs.DartDevice.Cpu(),
    ) -> V2Model:
        return cls(dartrs.DartV2Mixtral(hub_name, revision, dtype, device))


class MistralModel(V2Model):
    @classmethod
    def from_pretrained(
        cls,
        hub_name: str,
        revision: str | None = None,
        dtype: dartrs.DartDType = dartrs.DartDType.FP32,
        device: dartrs.DartDevice = dartrs.DartDevice.Cpu(),
    ) -> V2Model:
        return cls(dartrs.DartV2Mistral(hub_name, revision, dtype, device))



from dartrs.dartrs import (
    DartDevice,
    DartTokenizer,
)
from dartrs.v2 import (
    MixtralModel,
    GenerationConfig,
    compose_prompt,
    LengthTag,
    RatingTag,
    AspectRatioTag,
    IdentityTag,
)
from random import randint


def prepare_models():
    model = MixtralModel.from_pretrained("p1atdev/dart-v2-mixtral-160m-sft-2")
    tokenizer = DartTokenizer.from_pretrained("p1atdev/dart-v2-mixtral-160m-sft-2")

    return model, tokenizer


def test_generate():
    model, tokenizer = prepare_models()

    prompt = compose_prompt(
        prompt="1girl, cat ears",
    )
    config = GenerationConfig(
        prompt=prompt,
        tokenizer=tokenizer,
        seed=42,
    )

    output = model.generate(config)
    assert output is not None


def test_generate_different_seed():
    model, tokenizer = prepare_models()

    seeds = [1, 2900]
    results = []

    for seed in seeds:
        prompt = compose_prompt(
            prompt="1girl, cat ears",
        )
        config = GenerationConfig(
            prompt=prompt,
            tokenizer=tokenizer,
            seed=seed,
        )

        results.append(model.generate(config))

    assert results[0] != results[1]


def test_generate_random_seed():
    model, tokenizer = prepare_models()

    seeds = [randint(0, 10000) for _ in range(10)]
    results = []

    for seed in seeds:
        prompt = compose_prompt(
            prompt="1girl, cat ears",
        )
        config = GenerationConfig(
            prompt=prompt,
            tokenizer=tokenizer,
            seed=seed,
        )

        results.append(model.generate(config))

    for i in range(1, len(results)):
        assert results[i] != results[i - 1]


def test_generate_same_seed():
    model, tokenizer = prepare_models()

    seed = 12345

    prompt = compose_prompt(
        prompt="1girl, cat ears",
    )
    config = GenerationConfig(
        prompt=prompt,
        tokenizer=tokenizer,
        seed=seed,
    )

    output1 = model.generate(config)
    output2 = model.generate(config)

    assert output1 == output2


def test_generate_different_temperature():
    model, tokenizer = prepare_models()

    temperatures = [0.5, 1.0, 1.5]
    results = []

    for temperature in temperatures:
        prompt = compose_prompt(
            prompt="1girl, cat ears",
        )
        config = GenerationConfig(
            prompt=prompt,
            tokenizer=tokenizer,
            seed=42,
            temperature=temperature,
        )

        results.append(model.generate(config))

    assert results[0] != results[1]
    assert results[1] != results[2]
    assert results[0] != results[2]


def test_generate_different_top_p():
    model, tokenizer = prepare_models()

    top_ps = [0.5, 0.9, 0.99]
    results = []

    for top_p in top_ps:
        prompt = compose_prompt(
            prompt="1girl, cat ears",
        )
        config = GenerationConfig(
            prompt=prompt,
            tokenizer=tokenizer,
            seed=42,
            top_p=top_p,
        )

        results.append(model.generate(config))

    assert results[0] != results[1]
    assert results[1] != results[2]
    assert results[0] != results[2]


def test_generate_different_top_k():
    model, tokenizer = prepare_models()

    top_ks = [50, 100, 200]
    results = []

    for top_k in top_ks:
        prompt = compose_prompt(
            prompt="1girl, cat ears",
        )
        config = GenerationConfig(
            prompt=prompt,
            tokenizer=tokenizer,
            seed=42,
            top_k=top_k,
        )

        results.append(model.generate(config))

    assert results[0] != results[1]
    assert results[1] != results[2]
    assert results[0] != results[2]


def test_generate_different_prompt():
    model, tokenizer = prepare_models()

    prompts = [
        compose_prompt(
            prompt="1girl, cat ears",
        ),
        compose_prompt(
            prompt="2girls",
        ),
        compose_prompt(
            prompt="no humans, scenery",
        ),
    ]

    results = []

    for prompt in prompts:
        config = GenerationConfig(
            prompt=prompt,
            tokenizer=tokenizer,
            seed=42,
        )

        results.append(model.generate(config))

    assert results[0] != results[1]
    assert results[1] != results[2]



from abc import ABC

class DartDType:
    BF16: ...
    FP16: ...
    FP32: ...

class DartDevice:
    @classmethod
    def Cpu(cls) -> ...: ...
    @classmethod
    def Cuda(cls, id: int) -> ...: ...

class GenerationConfig:
    def __init__(
        self,
        device: DartDevice,
        tokenizer: DartTokenizer,
        prompt: str,
        eos_token: str | None = None,
        max_new_tokens: int | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        top_k: int | None = None,
        seed: int | None = None,
    ) -> None: ...

class DartV2Mistral:
    def __init__(
        self,
        hub_name: str,
        revision: str | None = None,
        dtype: DartDType = DartDType.FP32,
        device: DartDevice = DartDevice.Cpu(),
    ) -> None:
        raise NotImplementedError

    def generate(self, config: GenerationConfig) -> str:
        raise NotImplementedError

class DartV2Mixtral:
    def __init__(
        self,
        hub_name: str,
        revision: str | None = None,
        dtype: DartDType = DartDType.FP32,
        device: DartDevice = DartDevice.Cpu(),
    ) -> None:
        raise NotImplementedError

    def generate(self, config: GenerationConfig) -> str:
        raise NotImplementedError

class DartTokenizer:
    @staticmethod
    def from_pretrained(
        identifier,
        revision=...,
    ) -> DartTokenizer: ...

class SpecialTag(ABC):
    def __init__(self, tag: str) -> None: ...
    def to_tag(self) -> str: ...

class LengthTag(SpecialTag):
    VeryShort: LengthTag
    Short: LengthTag
    Medium: LengthTag
    Long: LengthTag
    VeryLong: LengthTag

class AspectRatioTag(SpecialTag):
    UltraWide: AspectRatioTag
    Wide: AspectRatioTag
    Square: AspectRatioTag
    Tall: AspectRatioTag
    UltraTall: AspectRatioTag

class RatingTag(SpecialTag):
    Sfw: RatingTag
    General: RatingTag
    Sensitive: RatingTag
    Nsfw: RatingTag
    Questionable: RatingTag
    Explicit: RatingTag

class IdentityTag(SpecialTag):
    Free: IdentityTag
    Lax: IdentityTag
    Strict: IdentityTag

def compose_prompt_v2(
    copyright: str,
    character: str,
    rating: RatingTag,
    aspect_ratio: AspectRatioTag,
    length: LengthTag,
    identity_level: IdentityTag,
    prompt: str,
) -> str: ...
